@@ -31,12 +31,16 @@
 #include "dr.h"
 #include "ofdt.h"
 #include "drmem.h"
+#include "common_numa.h"
 
-static int block_sz_bytes = 0;
+int block_sz_bytes = 0;
 static char *state_strs[] = {"offline", "online"};
 
 static char *usagestr = "-c mem {-a | -r} {-q <quantity> -p {variable_weight | ent_capacity} | {-q <quantity> | -s [<drc_name> | <drc_index>]}}";
 
+static struct ppcnuma_topology numa;
+static int numa_enabled = 0;
+
 /**
  * mem_usage
  * @brief return usage string
@@ -49,6 +53,21 @@ mem_usage(char **pusage)
 	*pusage = usagestr;
 }
 
+/**
+ * report_resource_count
+ * @brief Report the number of LMBs that were added or removed.
+ *
+ * Note that the format of this message is what is expected by the HMC,
+ * or other agent communicating through the RMC framework, and should not
+ * be changed.
+ *
+ * @param count number of LMBs
+ */
+static void report_resource_count(int count)
+{
+	printf("DR_TOTAL_RESOURCES=%d\n", count);
+}
+
 /**
  * get_phandle
  *
@@ -291,33 +310,82 @@ get_mem_node_lmbs(struct lmb_list_head *lmb_list)
 	return rc;
 }
 
+static int link_lmb_to_numa_node(struct dr_node *lmb)
+{
+	int nid;
+	struct ppcnuma_node *node;
+
+	nid = aa_index_to_node(&numa.aa, lmb->lmb_aa_index);
+	if (nid == -1)
+		return 0;
+
+	node = ppcnuma_fetch_node(&numa, nid);
+	if (!node)
+		return -ENOMEM;
+
+	lmb->lmb_numa_next = node->lmbs;
+	node->lmbs = lmb;
+	node->n_lmbs++;
+
+	if (node->n_cpus)
+		numa.lmb_count++;
+	else
+		numa.cpuless_lmb_count++;
+
+	return 0;
+}
+
+int add_lmb(struct lmb_list_head *lmb_list, uint32_t drc_index,
+	    uint64_t address, uint64_t lmb_sz, uint32_t aa_index,
+	    uint32_t flags)
+{
+	struct dr_node *lmb;
+
+	lmb = lmb_list_add(drc_index, lmb_list);
+	if (lmb == NULL) {
+		say(DEBUG, "Could not find LMB with drc-index of %x\n",
+		    drc_index);
+		return -1;
+	}
+
+	sprintf(lmb->ofdt_path, DYNAMIC_RECONFIG_MEM);
+	lmb->lmb_size = lmb_sz;
+	lmb->lmb_address = address;
+	lmb->lmb_aa_index = aa_index;
+
+	if (numa_enabled && link_lmb_to_numa_node(lmb))
+		return -ENOMEM;
+
+	if (flags & DRMEM_ASSIGNED) {
+		int rc;
+
+		lmb->is_owned = 1;
+
+		/* find the associated sysfs memory blocks */
+		rc = get_mem_scns(lmb);
+		if (rc)
+			return -1;
+	}
+
+	lmb_list->lmbs_found++;
+	return 0;
+}
 /**
- * get_dynamic_reconfig_lmbs
+ * get_dynamic_reconfig_lmbs_v1
  * @brief Retrieve lmbs from OF device tree located in the ibm,dynamic-memory
  * property.
  *
+ * @param lmb_sz size of LMBs
  * @param lmb_list pointer to lmb list head to populate
  * @returns 0 on success, !0 on failure
  */
 int
-get_dynamic_reconfig_lmbs(struct lmb_list_head *lmb_list)
+get_dynamic_reconfig_lmbs_v1(uint64_t lmb_sz, struct lmb_list_head *lmb_list)
 {
 	struct drconf_mem *drmem;
-	uint64_t lmb_sz;
 	int i, num_entries;
 	int rc = 0;
 
-	rc = get_property(DYNAMIC_RECONFIG_MEM, "ibm,lmb-size",
-			  &lmb_sz, sizeof(lmb_sz));
-
-	/* convert for LE systems */
-	lmb_sz = be64toh(lmb_sz);
-
-	if (rc) {
-		say(DEBUG, "Could not retrieve drconf LMB size\n");
-		return rc;
-	}
-
 	lmb_list->drconf_buf_sz = get_property_size(DYNAMIC_RECONFIG_MEM,
 						   "ibm,dynamic-memory");
 	lmb_list->drconf_buf = zalloc(lmb_list->drconf_buf_sz);
@@ -345,34 +413,123 @@ get_dynamic_reconfig_lmbs(struct lmb_list_head *lmb_list)
 	drmem = (struct drconf_mem *)
 				(lmb_list->drconf_buf + sizeof(num_entries));
 	for (i = 0; i < num_entries; i++) {
-		struct dr_node *lmb;
-
-		lmb = lmb_list_add(be32toh(drmem->drc_index), lmb_list);
-		if (lmb == NULL) {
-			say(DEBUG, "Could not find LMB with drc-index of %x\n",
-			    drmem->drc_index);
-			rc = -1;
+		rc = add_lmb(lmb_list, be32toh(drmem->drc_index),
+			     be64toh(drmem->address), lmb_sz,
+			     be32toh(drmem->assoc_index),
+			     be32toh(drmem->flags));
+		if (rc)
 			break;
-		}
 
-		sprintf(lmb->ofdt_path, DYNAMIC_RECONFIG_MEM);
-		lmb->lmb_size = lmb_sz;
-		lmb->lmb_address = be64toh(drmem->address);
-		lmb->lmb_aa_index = be32toh(drmem->assoc_index);
+		drmem++; /* trust your compiler */
+	}
 
-		if (be32toh(drmem->flags) & DRMEM_ASSIGNED) {
-			lmb->is_owned = 1;
+	return rc;
+}
 
-			/* find the associated sysfs memory blocks */
-			rc = get_mem_scns(lmb);
+/**
+ * get_dynamic_reconfig_lmbs_v2
+ * @brief Retrieve the LMBs from the ibm,dynamic-memory-v2 property
+ *
+ * @param lmb_sz LMB size
+ * @param lmb_list pointer to lmb_list head to populate
+ * @returns 0 on success, !0 on failure.
+ */
+int get_dynamic_reconfig_lmbs_v2(uint64_t lmb_sz,
+				 struct lmb_list_head *lmb_list)
+{
+	struct drconf_mem_v2 *drmem;
+	uint32_t lmb_sets;
+	int i, rc = 0;
+
+	lmb_list->drconf_buf_sz = get_property_size(DYNAMIC_RECONFIG_MEM,
+						   "ibm,dynamic-memory-v2");
+	lmb_list->drconf_buf = zalloc(lmb_list->drconf_buf_sz);
+	if (lmb_list->drconf_buf == NULL) {
+		say(DEBUG, "Could not allocate buffer to get dynamic "
+		    "reconfigurable memory\n");
+		return -1;
+	}
+
+	rc = get_property(DYNAMIC_RECONFIG_MEM, "ibm,dynamic-memory-v2",
+			  lmb_list->drconf_buf, lmb_list->drconf_buf_sz);
+	if (rc) {
+		say(DEBUG, "Could not retrieve dynamic reconfigurable memory "
+		    "property\n");
+		return -1;
+	}
+
+	/* The first integer of the buffer is the number of lmb sets */
+	lmb_sets = *(int *)lmb_list->drconf_buf;
+	lmb_sets = be32toh(lmb_sets);
+
+	/* Followed by the actual entries */
+	drmem = (struct drconf_mem_v2 *)
+				(lmb_list->drconf_buf + sizeof(lmb_sets));
+
+	for (i = 0; i < lmb_sets; i++) {
+		uint32_t drc_index, seq_lmbs;
+		uint64_t address;
+		int j;
+
+		address = be64toh(drmem->base_addr);
+		drc_index = be32toh(drmem->drc_index);
+		seq_lmbs = be32toh(drmem->seq_lmbs);
+
+		for (j = 0; j < seq_lmbs; j++) {
+			uint32_t aa_index = be32toh(drmem->aa_index);
+			uint32_t flags = be32toh(drmem->flags);
+
+			rc = add_lmb(lmb_list, drc_index, address,
+				     lmb_sz, aa_index, flags);
 			if (rc)
 				break;
+
+			drc_index++;
+			address += lmb_sz;
 		}
 
-		lmb_list->lmbs_found++;
 		drmem++; /* trust your compiler */
 	}
 
+	return rc;
+}
+
+/**
+ * get_dynamic_reconfig_lmbs
+ * @brief Retrieve lmbs from OF device tree located in the ibm,dynamic-memory
+ * property.
+ *
+ * @param lmb_list pointer to lmb list head to populate
+ * @returns 0 on success, !0 on failure
+ */
+int
+get_dynamic_reconfig_lmbs(struct lmb_list_head *lmb_list)
+{
+	struct stat sbuf;
+	uint64_t lmb_sz;
+	int rc = 0;
+
+	rc = get_property(DYNAMIC_RECONFIG_MEM, "ibm,lmb-size",
+			  &lmb_sz, sizeof(lmb_sz));
+
+	/* convert for LE systems */
+	lmb_sz = be64toh(lmb_sz);
+
+	if (rc) {
+		say(DEBUG, "Could not retrieve drconf LMB size\n");
+		return rc;
+	}
+
+	if (stat(DYNAMIC_RECONFIG_MEM_V1, &sbuf) == 0) {
+		rc = get_dynamic_reconfig_lmbs_v1(lmb_sz, lmb_list);
+	} else if ((read_dynamic_memory_v2 || numa_enabled) &&
+		   stat(DYNAMIC_RECONFIG_MEM_V2, &sbuf) == 0) {
+		rc = get_dynamic_reconfig_lmbs_v2(lmb_sz, lmb_list);
+	} else {
+		say(ERROR, "No dynamic reconfiguration LMBs found\n");
+		return -1;
+	}
+
 	say(INFO, "Found %d LMBs currently allocated\n", lmb_list->lmbs_found);
 	return rc;
 }
@@ -518,8 +675,7 @@ get_lmbs(unsigned int sort)
  * @param start_lmb first lmbs to be searched for an available lmb
  * @returns pointer to avaiable lmb on success, NULL otherwise
  */
-static struct dr_node *
-get_available_lmb(struct options *opts, struct dr_node *start_lmb)
+static struct dr_node *get_available_lmb(struct dr_node *start_lmb)
 {
 	uint32_t drc_index;
 	struct dr_node *lmb;
@@ -529,22 +685,21 @@ get_available_lmb(struct options *opts, struct dr_node *start_lmb)
 	for (lmb = start_lmb; lmb; lmb = lmb->next) {
 		int rc;
 
-		if (opts->usr_drc_name) {
-			drc_index = strtoul(opts->usr_drc_name, NULL, 0);
+		if (usr_drc_name) {
+			drc_index = strtoul(usr_drc_name, NULL, 0);
 
-			if ((strcmp(lmb->drc_name, opts->usr_drc_name))
+			if ((strcmp(lmb->drc_name, usr_drc_name))
 			    && (lmb->drc_index != drc_index))
 				continue;
-		} else if (opts->usr_drc_index) {
-			if (lmb->drc_index != opts->usr_drc_index)
+		} else if (usr_drc_index) {
+			if (lmb->drc_index != usr_drc_index)
 				continue;
 		}
 
 		if (lmb->unusable)
 			continue;
 
-		switch (opts->action) {
-		    case ADD:
+		if (usr_action == ADD) {
 			if (lmb->is_owned)
 				continue;
 
@@ -553,9 +708,7 @@ get_available_lmb(struct options *opts, struct dr_node *start_lmb)
 				continue;
 
 			usable_lmb = lmb;
-			break;
-
-		    case REMOVE:
+		} else if (usr_action == REMOVE) {
 			/* removable is ignored if AMS ballooning is active. */
 			if ((!balloon_active && !lmb->is_removable) ||
 			    (!lmb->is_owned))
@@ -820,7 +973,7 @@ add_device_tree_lmb(struct dr_node *lmb, struct lmb_list_head *lmb_list)
 static int
 get_mem_scn_state(struct mem_scn *mem_scn)
 {
-	char path[DR_PATH_MAX];
+	char path[PATH_MAX];
 	char state[8];
 	int file;
 	int rc;
@@ -860,10 +1013,11 @@ static int
 set_mem_scn_state(struct mem_scn *mem_scn, int state)
 {
 	int file;
-	char path[DR_PATH_MAX];
+	char path[PATH_MAX];
 	int rc = 0;
 	time_t t;
 	char tbuf[128];
+	int my_errno;
 
 	time(&t);
 	strftime(tbuf, 128, "%T", localtime(&t));
@@ -881,11 +1035,12 @@ set_mem_scn_state(struct mem_scn *mem_scn, int state)
 	}
 
 	rc = write(file, state_strs[state], strlen(state_strs[state]));
+	my_errno = errno;
 	close(file);
 
 	if (rc < 0) {
 		say(ERROR, "Could not write to %s to %s memory\n\t%s\n",
-		    path, state_strs[state], strerror(errno));
+		    path, state_strs[state], strerror(my_errno));
 		return rc;
 	}
 
@@ -921,10 +1076,11 @@ probe_lmb(struct dr_node *lmb)
 	int rc = 0;
 
 	probe_file = open(MEM_PROBE_FILE, O_WRONLY);
-	if (probe_file <= 0) {
+	if (probe_file == -1) {
+		int my_errno = errno;
 		say(DEBUG, "Could not open %s to probe memory\n",
 		    MEM_PROBE_FILE);
-		return errno;
+		return my_errno;
 	}
 
 	for (scn = lmb->lmb_mem_scns; scn; scn = scn->next) {
@@ -1016,19 +1172,18 @@ set_lmb_state(struct dr_node *lmb, int state)
  * @param lmb_list list of lmbs on the partition
  * @returns 0 on success, !0 otherwise
  */
-static int
-add_lmbs(struct options *opts, struct lmb_list_head *lmb_list)
+static int add_lmbs(struct lmb_list_head *lmb_list)
 {
 	int rc = 0;
 	struct dr_node *lmb_head = lmb_list->lmbs;
 	struct dr_node *lmb;
 
 	lmb_list->lmbs_modified = 0;
-	while (lmb_list->lmbs_modified < (int)opts->quantity) {
+	while (lmb_list->lmbs_modified < usr_drc_count) {
 		if (drmgr_timed_out())
 			break;
 
-		lmb = get_available_lmb(opts, lmb_head);
+		lmb = get_available_lmb(lmb_head);
 		if (lmb == NULL)
 			return -1;
 
@@ -1069,11 +1224,9 @@ add_lmbs(struct options *opts, struct lmb_list_head *lmb_list)
  * mem_add
  * @brief Add memory to the partition
  *
- * @param opts user options
  * @returns 0 on success, !0 otherwise
  */
-static int
-mem_add(struct options *opts)
+static int mem_add(void)
 {
 	struct lmb_list_head *lmb_list;
 	int rc;
@@ -1085,12 +1238,12 @@ mem_add(struct options *opts)
 		return -1;
 	}
 
-	say(DEBUG, "Attempting to add %d LMBs\n", opts->quantity);
-	rc = add_lmbs(opts, lmb_list);
+	say(DEBUG, "Attempting to add %d LMBs\n", usr_drc_count);
+	rc = add_lmbs(lmb_list);
 
 	say(DEBUG, "Added %d of %d requested LMB(s)\n", lmb_list->lmbs_modified,
-	    opts->quantity);
-	printf("DR_TOTAL_RESOURCES=%d\n", lmb_list->lmbs_modified);
+	    usr_drc_count);
+	report_resource_count(lmb_list->lmbs_modified);
 
 	free_lmbs(lmb_list);
 	return rc;
@@ -1100,22 +1253,20 @@ mem_add(struct options *opts)
  * remove_lmbs
  *
  * @param nr_lmbs
- * @param opts
  * @param lmb_list
  * @return 0 on success, !0 otherwise
  */
-static int
-remove_lmbs(struct options *opts, struct lmb_list_head *lmb_list)
+static int remove_lmbs(struct lmb_list_head *lmb_list)
 {
 	struct dr_node *lmb_head = lmb_list->lmbs;
 	struct dr_node *lmb;
 	int rc;
 
-	while (lmb_list->lmbs_modified < (int)opts->quantity) {
+	while (lmb_list->lmbs_modified < usr_drc_count) {
 		if (drmgr_timed_out())
 			break;
 
-		lmb = get_available_lmb(opts, lmb_head);
+		lmb = get_available_lmb(lmb_head);
 		if (!lmb)
 			return -1;
 
@@ -1161,16 +1312,13 @@ remove_lmbs(struct options *opts, struct lmb_list_head *lmb_list)
 /**
  * mem_remove
  *
- * @param opts
  * @return 0 on success, !0 otherwise
  */
-static int
-mem_remove(struct options *opts)
+static int mem_remove(void)
 {
 	struct lmb_list_head *lmb_list;
 	struct dr_node *lmb;
 	unsigned int removable = 0;
-	unsigned int requested = opts->quantity;
 	int rc = 0;
 
 	lmb_list = get_lmbs(LMB_RANDOM_SORT);
@@ -1198,24 +1346,24 @@ mem_remove(struct options *opts)
 			rc = -1;
 		}
 
-		if (removable < opts->quantity) {
+		if (removable < usr_drc_count) {
 			say(INFO, "Only %u LMBs are currently candidates "
 					"for removal.\n", removable);
-			opts->quantity = removable;
+			usr_drc_count = removable;
 		}
 	}
 
 	if (!rc) {
-		say(DEBUG, "Attempting removal of %d LMBs\n", opts->quantity);
-		rc = remove_lmbs(opts, lmb_list);
+		say(DEBUG, "Attempting removal of %d LMBs\n", usr_drc_count);
+		rc = remove_lmbs(lmb_list);
 	}
 
 	say(ERROR, "Removed %d of %d requested LMB(s)\n",
-	    lmb_list->lmbs_modified, requested);
-	if (lmb_list->lmbs_modified < requested)
+	    lmb_list->lmbs_modified, usr_drc_count);
+	if (lmb_list->lmbs_modified < usr_drc_count)
 		say(ERROR, "Unable to hotplug remove the remaining %d LMB(s)\n",
-		    requested - lmb_list->lmbs_modified);
-	printf("DR_TOTAL_RESOURCES=%d\n", lmb_list->lmbs_modified);
+		    usr_drc_count - lmb_list->lmbs_modified);
+	report_resource_count(lmb_list->lmbs_modified);
 
 	free_lmbs(lmb_list);
 	return rc;
@@ -1241,11 +1389,9 @@ mem_remove(struct options *opts)
  * of the ehea module.  If it is present we check its capabilities file
  * to determine if it can handle memory dlpar.
  *
- * @param action memory add/remove action that is to be checked for
  * @return 1 if we are ehea compatable, 0 otherwise
  */
-static int
-ehea_compatable(int action)
+static int ehea_compatable(void)
 {
 	FILE *fp;
 	uint64_t flags = 0;
@@ -1276,52 +1422,391 @@ ehea_compatable(int action)
 	/* Assume memory dlpar is not supported */
 	rc = 0;
 
-	if ((action == ADD) && (flags & MEM_ADD_ATTR))
+	if ((usr_action == ADD) && (flags & MEM_ADD_ATTR))
 		rc = 1;
 
-	if ((action == REMOVE) && (flags & MEM_RM_ATTR))
+	if ((usr_action == REMOVE) && (flags & MEM_RM_ATTR))
 		rc = 1;
 
 	if (!rc)
 		say(INFO, "The eHEA modules loaded on this system does not "
 		    "support memory DLPAR %s operations.\n",
-		    (action == ADD) ? "add" : "remove");
+		    (usr_action == ADD) ? "add" : "remove");
 	return rc;
 }
 
-int
-valid_mem_options(struct options *opts)
+int valid_mem_options(void)
 {
 	/* default to a quantity of 1 */
-	if (opts->quantity == 0)
-		opts->quantity = 1;
+	if (usr_drc_count == 0)
+		usr_drc_count = 1;
 
-	if ((opts->action != ADD) && (opts->action != REMOVE)) {
+	if ((usr_action != ADD) && (usr_action != REMOVE)) {
 		say(ERROR, "The '-r' or '-a' option must be specified for "
 		    "memory operations\n");
 		return -1;
 	}
 
 	/* The -s option can specify a drc name or drc index */
-	if (opts->usr_drc_name && !strncmp(opts->usr_drc_name, "0x", 2)) {
-		opts->usr_drc_index = strtoul(opts->usr_drc_name, NULL, 16);
-		opts->usr_drc_name = NULL;
+	if (usr_drc_name && !strncmp(usr_drc_name, "0x", 2)) {
+		usr_drc_index = strtoul(usr_drc_name, NULL, 16);
+		usr_drc_name = NULL;
 	}
 
 	return 0;
 }
 
-int
-drslot_chrp_mem(struct options *opts)
+static int remove_lmb_by_index(uint32_t drc_index)
+{
+	char cmdbuf[128];
+	int offset;
+
+	offset = sprintf(cmdbuf, "memory remove index 0x%x", drc_index);
+
+	return do_kernel_dlpar_common(cmdbuf, offset,
+				      1 /* Don't report error */);
+}
+
+static int remove_lmb_from_node(struct ppcnuma_node *node, uint32_t count)
+{
+	struct dr_node *lmb;
+	int err, done = 0, unlinked = 0;
+
+	say(DEBUG, "Try removing %d / %d LMBs from node %d\n",
+	    count, node->n_lmbs, node->node_id);
+
+	for (lmb = node->lmbs; lmb && done < count; lmb = lmb->lmb_numa_next) {
+		unlinked++;
+		err = remove_lmb_by_index(lmb->drc_index);
+		if (err)
+			say(WARN, "Can't remove LMB node:%d index:0x%x: %s\n",
+			    node->node_id, lmb->drc_index, strerror(-err));
+		else
+			done++;
+	}
+
+	/*
+	 * Decrement the node LMB's count since whatever is the success
+	 * of the removal operation, it will not be tried again on that
+	 * LMB.
+	 */
+	node->n_lmbs -= unlinked;
+
+	/*
+	 * Update the node's list of LMB to not process the one we removed or
+	 * tried to removed again.
+	 */
+	node->lmbs = lmb;
+
+	/* Update numa's counters */
+	if (node->n_cpus)
+		numa.lmb_count -= unlinked;
+	else
+		numa.cpuless_node_count -= unlinked;
+
+	if (!node->n_lmbs) {
+		node->ratio = 0; /* for sanity only */
+		if (node->n_cpus)
+			numa.cpu_count -= node->n_cpus;
+		else
+			numa.cpuless_node_count--;
+	}
+
+	say(INFO, "Removed %d LMBs from node %d\n", done, node->node_id);
+	return done;
+}
+
+#define min(a, b) ((a < b) ? a : b)
+
+static void update_cpuless_node_ratio(void)
+{
+	struct ppcnuma_node *node;
+	int nid;
+
+	/*
+	 * Assumptions:
+	 * 1. numa->cpuless_node_count is up to date
+	 * 2. numa->cpuless_lmb_count is up to date
+	 * Nodes with no memory and nodes with CPUs are ignored here.
+	 */
+	ppcnuma_foreach_node(&numa, nid, node) {
+		if (node->n_cpus || !node->n_lmbs)
+			continue;
+		node->ratio = (node->n_lmbs * 100) / numa.cpuless_lmb_count;
+	}
+}
+
+/*
+ * Remove LMBs from node without CPUs only.
+ * The more the node has LMBs, the more LMBs will be removed from it.
+ *
+ * We have to retry the operation multiple times because some LMB cannot be
+ * removed due to the page usage in the kernel. In that case, that LMB is no
+ * more taken in account and the node's LMB count is decremented, assuming that
+ * LMB is unremovable at this time. Thus each node's ratio has to be computed on
+ * each iteration. This is not a big deal, usually, there are not so much nodes.
+ */
+static int remove_cpuless_lmbs(uint32_t count)
+{
+	struct ppcnuma_node *node;
+	int nid;
+	uint32_t total = count, todo, done = 0, this_loop;
+
+	while (count) {
+		count = min(count, numa.cpuless_lmb_count);
+		if (!count)
+			break;
+
+		update_cpuless_node_ratio();
+
+		this_loop = 0;
+		ppcnuma_foreach_node(&numa, nid, node) {
+			if (!node->n_lmbs || node->n_cpus)
+				continue;
+
+			todo = (count * node->ratio) / 100;
+			todo = min(todo, node->n_lmbs);
+			/* Fix rounded value to 0 */
+			if (!todo && node->n_lmbs)
+				todo = (count - this_loop);
+
+			if (todo)
+				todo = remove_lmb_from_node(node, todo);
+
+			this_loop += todo;
+			done += todo;
+			if (done >= total)
+				break;
+		}
+
+		/* Don't continue if we didn't make any progress. */
+		if (!this_loop)
+			break;
+
+		count -= this_loop;
+	}
+
+	say(DEBUG, "%d / %d LMBs removed from the CPU less nodes\n",
+	    done, total);
+	return done;
+}
+
+static void update_node_ratio(void)
+{
+	int nid;
+	struct ppcnuma_node *node, *n, **p;
+	uint32_t cpu_ratio, mem_ratio;
+
+	/*
+	 * Assumptions:
+	 * 1. numa->cpu_count is up to date
+	 * 2. numa->lmb_count is up to date
+	 * Nodes with no memory and nodes with no CPU are ignored here.
+	 */
+
+	numa.ratio = NULL;
+	ppcnuma_foreach_node(&numa, nid, node) {
+		if (!node->n_lmbs || !node->n_cpus)
+			continue;
+		cpu_ratio = (node->n_cpus * 100) / numa.cpu_count;
+		mem_ratio = (node->n_lmbs * 100) / numa.lmb_count;
+
+		/* Say that CPU ratio is 90% of the ratio */
+		node->ratio = (cpu_ratio * 9 + mem_ratio) / 10;
+	}
+
+	/* Create an ordered link of the nodes */
+	ppcnuma_foreach_node(&numa, nid, node) {
+		if (!node->n_lmbs || !node->n_cpus)
+			continue;
+
+		p = &numa.ratio;
+		for (n = numa.ratio;
+		     n && n->ratio < node->ratio; n = n->ratio_next)
+			p = &n->ratio_next;
+		*p = node;
+		node->ratio_next = n;
+	}
+}
+
+/*
+ * Remove LMBs from node with CPUs.
+ *
+ * The less a node has CPU, the more memory will be removed from it.
+ *
+ * As for the CPU less nodes, we must iterate because some LMBs may not be
+ * removable at this time.
+ */
+static int remove_cpu_lmbs(uint32_t count)
+{
+	struct ppcnuma_node *node;
+	uint32_t total = count, todo, done = 0, this_loop;
+	uint32_t new_lmb_count;
+
+	while (count) {
+		count = min(count, numa.lmb_count);
+		if (!count)
+			break;
+
+		update_node_ratio();
+
+		new_lmb_count = numa.lmb_count - count;
+
+		this_loop = 0;
+		ppcnuma_foreach_node_by_ratio(&numa, node) {
+			if (!node->n_lmbs || !node->n_cpus)
+				continue;
+
+			todo = (new_lmb_count * node->ratio)  / 100;
+			todo = node->n_lmbs - min(todo, node->n_lmbs);
+			todo = min(count, todo);
+
+			if (todo) {
+				todo = remove_lmb_from_node(node, todo);
+				count -= todo;
+				this_loop += todo;
+			}
+
+			if (!count)
+				break;
+		}
+
+		/* Don't continue if we didn't make any progress. */
+		if (!this_loop)
+			break;
+		done += this_loop;
+	}
+
+	say(DEBUG, "%d / %d LMBs removed from the CPU nodes\n",
+	    done, total);
+	return done;
+}
+
+static void build_numa_topology(void)
+{
+	int rc;
+
+	rc = ppcnuma_get_topology(&numa);
+	if (rc)
+		return;
+
+	numa_enabled = 1;
+}
+
+static void clear_numa_lmb_links(void)
+{
+	int nid;
+	struct ppcnuma_node *node;
+
+	ppcnuma_foreach_node(&numa, nid, node)
+		node->lmbs = NULL;
+}
+
+static int numa_based_remove(uint32_t count)
+{
+	struct lmb_list_head *lmb_list;
+	struct ppcnuma_node *node;
+	int nid;
+	uint32_t done = 0;
+
+	/*
+	 * Read the LMBs
+	 * Link the LMBs to their node
+	 * Update global counter
+	 */
+	lmb_list = get_lmbs(LMB_NORMAL_SORT);
+	if (lmb_list == NULL) {
+		clear_numa_lmb_links();
+		return -1;
+	}
+
+	if (!numa.node_count) {
+		clear_numa_lmb_links();
+		free_lmbs(lmb_list);
+		return -EINVAL;
+	}
+
+	ppcnuma_foreach_node(&numa, nid, node) {
+		say(INFO, "node %4d %4d CPUs %8d LMBs\n",
+		    nid, node->n_cpus, node->n_lmbs);
+	}
+
+	done += remove_cpuless_lmbs(count);
+	count -= done;
+
+	done += remove_cpu_lmbs(count);
+
+	report_resource_count(done);
+
+	clear_numa_lmb_links();
+	free_lmbs(lmb_list);
+	return 0;
+}
+
+int do_mem_kernel_dlpar(void)
+{
+	char cmdbuf[128];
+	int rc, offset;
+
+
+	if (usr_action == REMOVE && usr_drc_count && !usr_drc_index) {
+		build_numa_topology();
+		if (numa_enabled) {
+			if (!numa_based_remove(usr_drc_count))
+				return 0;
+
+			/*
+			 * If the NUMA based removal failed, lets try the legacy
+			 * way.
+			 */
+			say(WARN, "Can't do NUMA based removal operation.\n");
+		}
+	}
+
+	offset = sprintf(cmdbuf, "%s ", "memory");
+
+	switch (usr_action) {
+	case ADD:
+		offset += sprintf(cmdbuf + offset, "add ");
+		break;
+	case REMOVE:
+		offset += sprintf(cmdbuf + offset, "remove ");
+		break;
+	default:
+		say(ERROR, "Invalid DRC type specified\n");
+		return -EINVAL;
+	}
+
+	if (usr_drc_index)
+		offset += sprintf(cmdbuf + offset, "index 0x%x",
+				  usr_drc_index);
+	else
+		offset += sprintf(cmdbuf + offset, "count %d", usr_drc_count);
+
+	rc = do_kernel_dlpar(cmdbuf, offset);
+
+	/* Memory DLPAR in the kernel is handled as an all-or-nothing
+	 * request, we either add all the requested LMBs or add none
+	 * of them.
+	 */
+	if (rc)
+		report_resource_count(0);
+	else
+		report_resource_count(usr_drc_count);
+
+	return rc;
+}
+
+int drslot_chrp_mem(void)
 {
 	int rc = -1;
 
-	if (opts->p_option) {
+	if (usr_p_option) {
 		/* This is a entitlement or weight change */
-		return update_sysparm(opts);
+		return update_sysparm();
 	}
 
-	if (! mem_dlpar_capable() || ! ehea_compatable(opts->action)) {
+	if (! mem_dlpar_capable() || ! ehea_compatable()) {
 		say(ERROR, "DLPAR memory operations are not supported on"
 		    "this kernel.");
 		return -1;
@@ -1330,17 +1815,16 @@ drslot_chrp_mem(struct options *opts)
 	/* The recursive nature of the routines that add/remove lmbs
 	 * require that the quantity be non-zero.
 	 */
-	if (opts->usr_drc_name)
-		opts->quantity = 1;
+	if (usr_drc_name)
+		usr_drc_count = 1;
 
-	switch (opts->action) {
-	    case ADD:
-		rc = mem_add(opts);
-		break;
-
-	    case REMOVE:
-		rc = mem_remove(opts);
-		break;
+	if (kernel_dlpar_exists()) {
+		rc = do_mem_kernel_dlpar();
+	} else {
+		if (usr_action == ADD)
+			rc = mem_add();
+		else if (usr_action == REMOVE)
+			rc = mem_remove();
 	}
 
 	return rc;
