@@ -35,7 +35,7 @@
 int output_level = 0;
 int log_fd = 0;
 
-int is_lsslot_cmd = 1;
+int read_dynamic_memory_v2 = 1;
 
 extern int lsslot_chrp_cpu(void);
 
